#include <Arduino.h>
#include <Servo.h>
#define pin_base 3   // Servo Base
#define pin_L1 5     // Servo Hombro
#define pin_L2 6     // Servo Codo
#define solenoide 4  //(solenoide) Sensor Ultrasonico
//#define sensor 2 //Interrupciónencoder
#define led 12 //aviso arduino
String inputString = "";
bool stringComplete = false;
// Banda trasportadora
int IN3= 9;  //Sentido de giro
int IN4 = 10;  //Sentido de giro
int ENB = 11; //PWM
int speedMotor = 100; //Velocidad ideal 140

//Coordinación banda brazo
unsigned long tiempoInicial=0;
unsigned long tiempoActual=0;
unsigned long intervaloEspera = 4380; //  segundos en milisegundos4880
bool coordinacion=false;

//Detecciones de la cámara sobre el engranaje
int diametro;  
int y_pos ;
int n_dientes ;

//Movimiento del brazo 
int steps=90;  //Pasos por movimientos 
int pausa=5;    //Lentitud del paso

//Servos
Servo base; //Servo base
Servo L1;   //Servo Q1
Servo L2;   //Servo Q2
int estado = 0;         //Comienza en el estado 0

//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------- void setup -----------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void setup()
{
  Serial.begin(9600); //Se declara el canal de comunicación, los baudios son 9600
  
  // Servos y sus conexiones
  base.attach(pin_base); 
  L1.attach(pin_L1);
  L2.attach(pin_L2);
  
  // Solenoide
  pinMode(led,OUTPUT);
  pinMode (solenoide, OUTPUT);

  //En caso de conectar el encoder
  //attachInterrupt(digitalPinToInterrupt(sensor), interrupcion, RISING); //interrupcion por flanco de subida pin 2

  // Banda transportadora
  pinMode (ENB, OUTPUT);    // Input4 conectada al pin 3 pin de velocidad 
  pinMode (IN3, OUTPUT);    // Input3 conectada al pin 9
  pinMode (IN4, OUTPUT);    // Input3 conectada al pin 10
  
 
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//--------------------- void loop ----------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void loop()
{
  switch (estado) {
    //--------------------------------------------------------//
    //--------------------------------------------------------//
    // Estado 0: Se activa la banda, se va a la posición de home, y el brazo se coloca cerca de la banda
    case 0:

      //Motor gira en un sentido IZQ-DER CAMARA-BRAZO(Activación de banda)
      analogWrite (ENB, speedMotor); //speed motor

      //Controlar el sentido de giro
      digitalWrite (IN3, LOW);       
      digitalWrite (IN4, HIGH); 

      //Brazo se ubica en el home
      axis_home();
      delay(100);

      //Ubicación de espera sobre la banda
      inversa(245,0,50);
      estado = 1;

      //Lectura de datos enviados desde la cámara
      serialEvent();
      serialReading();
      
      
      break;
      
    //---------------------------------------------------------------------------------
    //----------------- Detectando un objeto, posición de recolección -----------------
    //La camara detecta el objeto y declara N de dientes y posición de engrane en eje "y", el
    //solenoide se activa, inversa(sobre banda, camara, sobre banda)
    case 1:
      //Solo para identificar que se encuentra en el estado 1 y que termino la lectura de datos
      //Serial.println("case1");

      //Enciende el solenoide y la led para indicar que esta esperando al engranaje
      digitalWrite(solenoide,LOW);
      digitalWrite(led,HIGH);
      delay(intervaloEspera);

      //Tras el envio de posición enviado por la cámara el brazo se mueve a la altura del engrane y a su posición en y
      inversa(210+y_pos,0,38); 
      delay(2000); 
      estado = 2;
      break;
      
    //--------------------------------------------------------  
    //----------------- LLevar a cajas el objeto -----------------
    // Se declaran posiciones de caja, función if
    

    case 2:

    //Ya con el engranaje en el solenoide se pasa a la clasificación y traslado a cajas
    //Imprime el # de dientes para poder ver si existe algún error en la clasificación
    //  Serial.println(n_dientes);

    //Clasificación de dientes con sus correspondientes posiciones de cajas 
    //Se describe solo la de 16 dientes, para las demás se sigue la misma lógica
      if (n_dientes==16)
      {
        // Se debe levantar el brazo para evitar raspar la cinta y que se suelte el engrane
        inversa(230,0,70);
        //Posición de la caja
        inversa(150,70,90);
        digitalWrite(solenoide,HIGH);// se suelta el engrane
        estado =0; //Se vuelve al home para seguir con otro engrane más 
        break;
      }
      if (n_dientes==14)
      {
        inversa(230,0,70);
        inversa(150,0,90);
        digitalWrite(solenoide,HIGH);// se suelta el engrane
        estado =0;
        break;
      }
      if (n_dientes==12)
      {
        inversa(230,0,70);
        inversa(150,-70,90);
        digitalWrite(solenoide,HIGH);// se suelta el engrane 
        estado = 0;
        break;
      }
      // En caso de que no exista o no se logre identificar un objeto es llevado afuera de las cajas 
      else{
        inversa(230,0,70);
        inversa(100,-70,90);
        digitalWrite(solenoide,HIGH);// se suelta el engrane 
        estado = 0;
        break;
      }
      break;
  }
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------ Control Servos a Posicion Base ----------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void axis_home()
{
  inversa(150 , 0, 150); // Posicion Home
  delay(100);
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//-------------- Cinematica Inversa --------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void inversa(float x1, float y1, float z1)
{
  // Largo de los links del brazo
  float l1 = 142; // Largo entre el eje del codo y del hombro
  float l2 = 150; // Largo del antebrazo
  float l3 = 36;  // Largo entre el fin del antebrazo y el eje de la muñecar
  float l5 = 35;  // Altura del gripper segun donde se quiera agarrar el objeto (Hacerlo variar en caso de que se quiera variar mucho el ancho del objeto agarrado) 
  float z0 = 65;  // Altura desde el piso hasta el servo del Hombro

  // Desfase de X y Z dadas las dimensiones del gripper en caso de que se quiera agarrar siempre perpendicular a la banda 
  float x = x1; // es -0
  float y = y1;  // es -0
  float z = z1 +l5; //+l5

  // Calculo de la inversa del brazo
  float q0_rad = atan(y / x);
  float q2_rad = acos((pow((l1 * cos(q0_rad)), 2) + pow((l2 * cos(q0_rad)), 2) - pow((x - l3 * cos(q0_rad)), 2) - pow((z - z0), 2)) / (2 * l1 * l2 * pow((cos(q0_rad)), 2)));
  float q1_rad = atan((x - l3 * cos(q0_rad)) / (z - z0)) - acos((pow((l1 * cos(q0_rad)), 2) - pow((l2 * cos(q0_rad)), 2) + pow((x - l3 * cos(q0_rad)), 2) + pow((z - z0), 2)) / (2 * l1 * cos(q0_rad) * sqrt(pow((x - l3 * cos(q0_rad)), 2) + pow((z - z0), 2))));
  float q3_rad = 3.14 / 2 + q0_rad;

  // Paso desde radianes a grados de los angulos 
  float q0 = q0_rad * 4068 / 71;
  float q1 = q1_rad * 4068 / 71;
  float q2 = q2_rad * 4068 / 71;

  // Mover los servos según los qi calculados para la configuracion de los servos
  move_axis( 1, 90 - 2 * q0 );
  move_axis( 2, 90 + q1);
  move_axis( 3, q2 - q1);
 
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//-------------- Cinematica Directa --------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
// Funcion si es que desease monitorear las posiciones a las que estaria llegando el robot 
void directa(int q1, int q2, int q3, int q4, int q5)
{
  
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//-------------- Control Servos ------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void move_axis(int servoId, float Meta)
{ //Grados en los que se encuentra cada servo:
  int gradosB=base.read();   
  int gradosL1=L1.read();
  int gradosL2=L2.read();

//División de los movimientos en base a los parámetros de pausa y pasos
//Los 3 servos siguen la misma logíca
  if (servoId == 1) // Servo Base
  { for (int i = 0; i <= steps; i++) {
    int currentPos = gradosB + ((Meta- gradosB) * i) / steps; //Se divide el movimiento del servo en N pasos(step)
    base.write(currentPos); // Mueve el servo a la posición actual
    delay(pausa); //Pausa para ralentizar el movimiento 
  }
  //delay(100);
  }
  //--------------------------------------------------------
  if (servoId == 2) // Servo Hombro
  { for (int i = 0; i <= steps; i++) {
    int currentPos = gradosL1 + ((Meta- gradosL1) * i) / steps;
    L1.write(currentPos); // Mueve el servo a la posición actual
    delay(pausa);
  }
  //delay(100);
  }
  //--------------------------------------------------------
  if (servoId == 3) // Servo Codo
  { for (int i = 0; i <= steps; i++) {
    int currentPos = gradosL2 + ((Meta- gradosL2) * i) / steps;
    L2.write(currentPos); // Mueve el servo a la posición actual
    delay(pausa);
  }
  //delay(100);
  }}

void serialReading() {
  if (stringComplete) {
    digitalWrite(led, HIGH);
    Serial.println(inputString);
    // Divida readString por un delimitador predefinido de una manera simple. '%'(porcentaje) se define como el delimitador en este proyecto.
    int delimiter, delimiter_1, delimiter_2, delimiter_3;
    delimiter = inputString.indexOf(",");
    delimiter_1 = inputString.indexOf(",", delimiter + 1);
    delimiter_2 = inputString.indexOf(",", delimiter_1 +1);
    delimiter_3 = inputString.indexOf(",", delimiter_2 +1);

    // Defina las variables que se ejecutarán en el código más adelante mediante la recopilación de información de readString como subcadenas.

    String first = inputString.substring(delimiter + 1, delimiter_1);
    String second = inputString.substring(delimiter_1 + 1, delimiter_2);
    String third = inputString.substring(delimiter_2 + 1, delimiter_3);

    diametro = first.toInt();
    y_pos = second.toInt();
    n_dientes = third.toInt();

  
      digitalWrite(led, LOW);
      Serial.print(diametro);
    Serial.print(" ");
    
    Serial.print(y_pos);
    Serial.print(" ");
    Serial.println(n_dientes);


    // Limpiar la cadena
    inputString = "";
    stringComplete = false;
  }

}
void serialEvent()
{
  while (!stringComplete) {
    if(Serial.available())
    {
    // obtener el nuevo byte
    char inChar = (char)Serial.read();
    // Añadir esto al inputString:
    inputString += inChar;
  // si el carácter entrante es una nueva línea, establezca una bandera para que el ciclo principal pueda
  // hacer algo al respecto
    if (inChar == '\n') {
      stringComplete = true;
      
    }
    }
  }
}